import arcpy
arcpy.env.overwriteOutput = True

import datetime
import netCDF4 as nc
import numpy as np
import os

import random

class Toolbox(object):
    def __init__(self):
        """Define the toolbox (the name of the toolbox is the name of the
        .pyt file)."""
        self.label = "NCEP Toolbox"
        self.alias = ""

        # List of tool classes associated with this toolbox
        self.tools = [NcToRaster, CreateProfile]

class NcToRaster(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "NCEP Reanalysis NetCDF to Raster Layer"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        #Define parameter definitions  
        param0 = arcpy.Parameter(
            displayName="Input NCEP Reanalysis NetCDF",
            name="in_netcdf",
            datatype="DEFile",
            parameterType="Required",
            direction="Input")       
        param1 = arcpy.Parameter(
            displayName="Variable Name",
            name="in_variable",
            datatype="String",
            parameterType="Required",
            direction="Input")      
        param2 = arcpy.Parameter(
            displayName="Index of Time",
            name="in_date",
            datatype="String",
            parameterType="Required",
            direction="Input")
        param3 = arcpy.Parameter(
            name="in_pressure_index", 
            displayName="Index of Pressure Level", 
            direction="Input", 
            datatype="Long",
            parameterType="Required")
        param4 = arcpy.Parameter(
            name="out_raster",
            displayName="Save raster as",
            direction="Output",
            datatype="DERasterDataset",
            parameterType="Required"
        )
    
        params = [param0, param1, param2, param3, param4]
    
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    #################################################################
    ## We transformed codes used in Python console to a function   ##
    ## so we can make a tool! Then we can re-use our codes easily! ##
    #################################################################
    def execute(self, parameters, messages):
        """The source code of the tool."""
        ## Compare codes written here and which in Python console ##
        def converting(path, varName, timeIndex, levelIndex):
            ncfile = nc.Dataset(path)
            var = ncfile.variables[varName]
            tempRaster = arcpy.NumPyArrayToRaster(var[timeIndex,levelIndex,:,:], arcpy.Point(0,-90), 2.5, 2.5)            
            return tempRaster
            
        # Supporting codes to run python codes as a tool in ArcGIS   
        in_netcdf = parameters[0].valueAsText
        in_var = parameters[1].valueAsText   
        in_time = parameters[2].valueAsText
        in_level = parameters[3].valueAsText
        out_raster = parameters[4].valueAsText
        timeIndex = int(in_time)
        levelIndex = int(in_level)
        
        # Let's do a debug, show a message to display date
        try:
            p = os.path.basename(in_netcdf)
            y = int(p.split('.')[1])
            s = datetime.datetime(y,1,1,0,0,0) + datetime.timedelta(seconds=timeIndex*6*3600)
            arcpy.AddMessage("Time at index of %d is %s" % (timeIndex, s))
        except:
            pass
        # Call the converting function above.
        tempRaster = converting(in_netcdf, in_var, timeIndex, levelIndex)
        preProjectName = 'temp_' + str(random.randint(10000,99999))
        tempRaster.save(preProjectName)
        # In-place projection fixes longitude shifting problem as well as adds spatial reference for output raster.
        arcpy.DefineProjection_management(preProjectName, 
                                          "GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]")
        arcpy.ProjectRaster_management(in_raster=preProjectName, 
                                       in_coor_system="GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]",
                                       out_raster=out_raster,
                                       out_coor_system="GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]]")
        arcpy.Delete_management(preProjectName)
        return

####################################################################
## This is the tool for Part 3 of the lab                         ##
####################################################################

class CreateProfile(object):
    def __init__(self):
        """Define the tool (tool name is the name of the class)."""
        self.label = "Create Vertial Profile"
        self.description = ""
        self.canRunInBackground = False

    def getParameterInfo(self):
        """Define parameter definitions"""
        #Define parameter definitions  
        param0 = arcpy.Parameter(
            displayName="Input NCEP Reanalysis NetCDF",
            name="in_netcdf",
            datatype="DEFile",
            parameterType="Required",
            direction="Input"
        )       
        param1 = arcpy.Parameter(
            displayName="Variable Name",
            name="in_variable",
            datatype="String",
            parameterType="Required",
            direction="Input"
        )  
        param1_1 = arcpy.Parameter(
            displayName="Geopotential Height NetCDF",
            name="in_hgt",
            datatype="DEFile",
            parameterType="Required",
            direction="Input"
        )
        param2 = arcpy.Parameter(
            displayName="Index of Time",
            name="in_date",
            datatype="Long",
            parameterType="Required",
            direction="Input"
        )
        param3 = arcpy.Parameter(
            name="in_cut_line", 
            displayName="Line Feature as Cutting Line", 
            direction="Input", 
            datatype="GPFeatureLayer",
            parameterType="Required"
        )
        param4 = arcpy.Parameter(
            name="out_feature",
            displayName="Save Sectional Plane As",
            direction="Output",
            datatype="DEFeatureClass",
            parameterType="Required"
        )
    
        params = [param0, param1, param1_1, param2, param3, param4]
    
        return params

    def isLicensed(self):
        """Set whether tool is licensed to execute."""
        return True

    def updateParameters(self, parameters):
        """Modify the values and properties of parameters before internal
        validation is performed.  This method is called whenever a parameter
        has been changed."""
        return

    def updateMessages(self, parameters):
        """Modify the messages created by internal validation for each tool
        parameter.  This method is called after internal validation."""
        return

    def execute(self, parameters, messages):
        """The source code of the tool."""
        
        # One line to unpack them in order
        in_netcdf, in_var, in_hgt, in_time, in_cut, out_plane = [p.valueAsText for p in parameters]
        
        air = nc.Dataset(in_netcdf).variables['air'][int(in_time),:,:,:]
        hgt = nc.Dataset(in_hgt).variables['hgt'][int(in_time),:,:,:]
        p = nc.Dataset(in_netcdf).variables['level'][:]
        
        # Read start point and end point of cutline
        # Remember to shift longitude+360
        # Remember to left-corner is 0,0
        c = arcpy.SearchCursor(in_cut)
        
        for row in c:
            feat = row.getValue('SHAPE')
            line = feat.getPart()
            x0 = line[0][0].X
            y0 = line[0][0].Y
            x1 = line[0][1].X
            y1 = line[0][1].Y
            
            length = int(np.hypot(x1-x0, y1-y0))
            x,y = np.linspace(x0,x1,length), np.linspace(y0,y1,length)
            xi = (x + 360) / 2.5
            yi = (90 - y) / 2.5       
            
            #Just in case there is more than one segment in cut line
            break

        result = []
        
        #Extract them layer by layer
        for z in range(air.shape[0]):
            l = air[z,:,:]
            h = hgt[z,:,:]
            ti = l[yi.astype(np.int), xi.astype(np.int)]
            hi = h[yi.astype(np.int), xi.astype(np.int)]
            # Convert to alt with factor
            hi = hi / 1000
            for j in range(len(ti)):
                #Add to result collect
                result.append(((x[j], y[j], hi[j]), ti[j]))
        
        # Put into structured array
        myarray = np.array(result, np.dtype([('XYZ', '<f8', 3), ('AIR', '<f8')]))      
        arcpy.da.NumPyArrayToFeatureClass(myarray, out_plane, ['XYZ'], arcpy.SpatialReference(4326))
        return
